@@ -4,10 +4,8 @@ __conf = {
     "gradient_normalization": "path",
     "spherical_harmonics_algorithm": "automatic",
     "spherical_harmonics_normalization": "component",
-    "specialized_code": False,
-    "optimize_einsums": True,
-    "custom_einsum_vjp": False,
-    "fuse_all": False,
+    "custom_einsum_jvp": False,
+    "fused": False,
 }
 
 
