@@ -4,7 +4,7 @@ import collections
 import itertools
 from functools import lru_cache, partial
 from math import sqrt
-from typing import Any, List, Optional, Union
+from typing import Callable, List, Optional, Tuple, Union
 
 import jax
 import jax.numpy as jnp
@@ -46,10 +46,10 @@ class FunctionalTensorProduct:
 
     def __init__(
         self,
-        irreps_in1: Any,
-        irreps_in2: Any,
-        irreps_out: Any,
-        instructions: List[Any],
+        irreps_in1: Irreps,
+        irreps_in2: Irreps,
+        irreps_out: Irreps,
+        instructions: List[Tuple[int, int, int, str, bool, Optional[float]]],
         in1_var: Optional[List[float]] = None,
         in2_var: Optional[List[float]] = None,
         out_var: Optional[List[float]] = None,
@@ -132,14 +132,12 @@ class FunctionalTensorProduct:
 
     def left_right(
         self,
-        weights: List[jnp.ndarray],
+        weights: Union[List[jnp.ndarray], jnp.ndarray],
         input1: IrrepsArray,
         input2: IrrepsArray = None,
         *,
-        specialized_code=None,
-        optimize_einsums=None,
-        custom_einsum_vjp=None,
-        fuse_all=None,
+        custom_einsum_jvp: bool = None,
+        fused: bool = None,
     ) -> IrrepsArray:
         r"""Compute the tensor product of two input tensors.
 
@@ -147,40 +145,27 @@ class FunctionalTensorProduct:
             weights (array or list of arrays): The weights of the tensor product.
             input1 (IrrepsArray): The first input tensor.
             input2 (IrrepsArray): The second input tensor.
-            specialized_code (bool): If True, use the specialized code for the
-                tensor product.
-            optimize_einsums (bool): If True, optimize the einsum code.
-            custom_einsum_vjp (bool): If True, use the custom vjp for the einsum
-                code.
-            fuse_all (bool): If True, fuse all the einsums.
+            custom_einsum_jvp (bool): If True, use the custom jvp for the einsum code.
+            fused (bool): If True, fuse all the einsums.
 
         Returns:
             `IrrepsArray`: The output tensor.
         """
-        if specialized_code is None:
-            specialized_code = config("specialized_code")
-        if optimize_einsums is None:
-            optimize_einsums = config("optimize_einsums")
-        if custom_einsum_vjp is None:
-            custom_einsum_vjp = config("custom_einsum_vjp")
-        if fuse_all is None:
-            fuse_all = config("fuse_all")
+        if custom_einsum_jvp is None:
+            custom_einsum_jvp = config("custom_einsum_jvp")
+        if fused is None:
+            fused = config("fused")
 
         if input2 is None:
             weights, input1, input2 = [], weights, input1
 
-        input1 = IrrepsArray.from_any(self.irreps_in1, input1)
-        input2 = IrrepsArray.from_any(self.irreps_in2, input2)
-
         return _left_right(
             self,
             weights,
-            input1,
-            input2,
-            specialized_code=specialized_code,
-            optimize_einsums=optimize_einsums,
-            custom_einsum_vjp=custom_einsum_vjp,
-            fuse_all=fuse_all,
+            input1.convert(self.irreps_in1),
+            input2.convert(self.irreps_in2),
+            custom_einsum_jvp=custom_einsum_jvp,
+            fused=fused,
         )
 
     def right(
@@ -188,41 +173,29 @@ class FunctionalTensorProduct:
         weights: List[jnp.ndarray],
         input2: IrrepsArray = None,
         *,
-        specialized_code=None,
-        optimize_einsums=None,
-        custom_einsum_vjp=None,
-        fuse_all=None,
+        custom_einsum_jvp=None,
     ) -> jnp.ndarray:
         r"""Compute the right contraction of the tensor product.
 
         Args:
             weights (array or list of arrays): The weights of the tensor product.
             input2 (IrrepsArray): The second input tensor.
-            optimize_einsums (bool): If True, optimize the einsum code.
-            custom_einsum_vjp (bool): If True, use the custom vjp for the einsum code.
+            custom_einsum_jvp (bool): If True, use the custom jvp for the einsum code.
 
         Returns:
             A matrix of shape ``(irreps_in1.dim, irreps_out.dim)``.
         """
-        if specialized_code is None:
-            specialized_code = config("specialized_code")
-        if optimize_einsums is None:
-            optimize_einsums = config("optimize_einsums")
-        if custom_einsum_vjp is None:
-            custom_einsum_vjp = config("custom_einsum_vjp")
-        if fuse_all is None:
-            fuse_all = config("fuse_all")
+        if custom_einsum_jvp is None:
+            custom_einsum_jvp = config("custom_einsum_jvp")
 
         if input2 is None:
             weights, input2 = [], weights
 
-        input2 = IrrepsArray.from_any(self.irreps_in2, input2)
         return _right(
             self,
             weights,
-            input2,
-            optimize_einsums=optimize_einsums,
-            custom_einsum_vjp=custom_einsum_vjp,
+            input2.convert(self.irreps_in2),
+            custom_einsum_jvp=custom_einsum_jvp,
         )
 
     def __repr__(self):
@@ -320,31 +293,21 @@ def _normalize_instruction_path_weights(
     return [update(instruction) for instruction in instructions]
 
 
-@partial(
-    jax.jit, static_argnums=(0,), static_argnames=("specialized_code", "optimize_einsums", "custom_einsum_vjp", "fuse_all")
-)
+@partial(jax.jit, static_argnums=(0,), static_argnames=("custom_einsum_jvp", "fused"))
 @partial(jax.profiler.annotate_function, name="TensorProduct.left_right")
 def _left_right(
     self: FunctionalTensorProduct,
-    weights,
-    input1,
-    input2,
+    weights: Union[List[jnp.ndarray], jnp.ndarray],
+    input1: IrrepsArray,
+    input2: IrrepsArray,
     *,
-    specialized_code=False,
-    optimize_einsums=True,
-    custom_einsum_vjp=False,
-    fuse_all=False,
+    custom_einsum_jvp: bool = False,
+    fused: bool = False,
 ):
-
-    # = Short-circut for zero dimensional =
     if self.irreps_in1.dim == 0 or self.irreps_in2.dim == 0 or self.irreps_out.dim == 0:
         return IrrepsArray.zeros(self.irreps_out, ())
 
-    if custom_einsum_vjp:
-        assert optimize_einsums
-        einsum = opt_einsum
-    else:
-        einsum = partial(jnp.einsum, optimize="optimal" if optimize_einsums else "greedy")
+    einsum = opt_einsum if custom_einsum_jvp else jnp.einsum
 
     if isinstance(weights, list):
         assert len(weights) == len([ins for ins in self.instructions if ins.has_weight]), (
@@ -368,85 +331,19 @@ def _left_right(
     assert input1.ndim == 1, f"input1 is shape {input1.shape}. Execting ndim to be 1. Use jax.vmap to map over input1"
     assert input2.ndim == 1, f"input2 is shape {input2.shape}. Execting ndim to be 1. Use jax.vmap to map over input2"
 
-    if fuse_all:
-        with jax.ensure_compile_time_eval():
-            num_path = weights_flat.size
-            has_path_with_no_weights = any(not ins.has_weight for ins in self.instructions)
-            i = 0
+    if fused:
+        return _fused_left_right(self, weights_flat, input1, input2, einsum)
+    else:
+        return _block_left_right(self, weights_list, input1, input2, einsum)
 
-            if has_path_with_no_weights:
-                num_path += 1
-                i += 1
-
-            big_w3j = jnp.zeros(
-                (
-                    num_path,
-                    self.irreps_in1.dim,
-                    self.irreps_in2.dim,
-                    self.irreps_out.dim,
-                )
-            )
-            for ins in self.instructions:
-                mul_ir_in1 = self.irreps_in1[ins.i_in1]
-                mul_ir_in2 = self.irreps_in2[ins.i_in2]
-                mul_ir_out = self.irreps_out[ins.i_out]
-                m1, m2, mo = mul_ir_in1.mul, mul_ir_in2.mul, mul_ir_out.mul
-                d1, d2, do = mul_ir_in1.ir.dim, mul_ir_in2.ir.dim, mul_ir_out.ir.dim
-                s1 = self.irreps_in1[: ins.i_in1].dim
-                s2 = self.irreps_in2[: ins.i_in2].dim
-                so = self.irreps_out[: ins.i_out].dim
-
-                w3j = clebsch_gordan(mul_ir_in1.ir.l, mul_ir_in2.ir.l, mul_ir_out.ir.l)
-
-                def set_w3j(i, u, v, w):
-                    return big_w3j.at[
-                        i,
-                        s1 + u * d1 : s1 + (u + 1) * d1,
-                        s2 + v * d2 : s2 + (v + 1) * d2,
-                        so + w * do : so + (w + 1) * do,
-                    ].add(ins.path_weight * w3j)
-
-                if ins.connection_mode == "uvw":
-                    assert ins.has_weight
-                    for u, v, w in itertools.product(range(m1), range(m2), range(mo)):
-                        big_w3j = set_w3j(i, u, v, w)
-                        i += 1
-                elif ins.connection_mode == "uvu":
-                    assert ins.has_weight
-                    for u, v in itertools.product(range(m1), range(m2)):
-                        big_w3j = set_w3j(i, u, v, u)
-                        i += 1
-                elif ins.connection_mode == "uvv":
-                    assert ins.has_weight
-                    for u, v in itertools.product(range(m1), range(m2)):
-                        big_w3j = set_w3j(i, u, v, v)
-                        i += 1
-                elif ins.connection_mode == "uuu":
-                    for u in range(m1):
-                        if ins.has_weight:
-                            big_w3j = set_w3j(i, u, u, u)
-                            i += 1
-                        else:
-                            big_w3j = set_w3j(0, u, u, u)
-                else:
-                    assert False
-
-        if has_path_with_no_weights and big_w3j.shape[0] == 1:
-            big_w3j = big_w3j.reshape(big_w3j.shape[1:])
-            out = einsum("ijk,i,j->k", big_w3j, input1.array, input2.array)
-        else:
-            if has_path_with_no_weights:
-                weights_flat = jnp.concatenate([jnp.ones((1,)), weights_flat])
-
-            out = einsum(
-                "p,pijk,i,j->k",
-                weights_flat,
-                big_w3j,
-                input1.array,
-                input2.array,
-            )
-        return IrrepsArray(self.irreps_out, out)
 
+def _block_left_right(
+    self: FunctionalTensorProduct,
+    weights_list: List[jnp.ndarray],
+    input1: IrrepsArray,
+    input2: IrrepsArray,
+    einsum: Callable,
+) -> IrrepsArray:
     @lru_cache(maxsize=None)
     def multiply(in1, in2, mode):
         if mode == "uv":
@@ -475,6 +372,7 @@ def _left_right(
 
         x1 = input1.list[ins.i_in1]
         x2 = input2.list[ins.i_in2]
+
         if x1 is None or x2 is None:
             out_list += [None]
             continue
@@ -487,67 +385,18 @@ def _left_right(
 
         if ins.connection_mode == "uvw":
             assert ins.has_weight
-            if specialized_code and (
-                mul_ir_in1.ir.l,
-                mul_ir_in2.ir.l,
-                mul_ir_out.ir.l,
-            ) == (0, 0, 0):
-                out = ins.path_weight * einsum("uvw,uv->w", w, xx.reshape(mul_ir_in1.dim, mul_ir_in2.dim))
-            elif specialized_code and mul_ir_in1.ir.l == 0:
-                out = ins.path_weight * einsum("uvw,u,vj->wj", w, x1.reshape(mul_ir_in1.dim), x2) / sqrt(mul_ir_out.ir.dim)
-            elif specialized_code and mul_ir_in2.ir.l == 0:
-                out = ins.path_weight * einsum("uvw,ui,v->wi", w, x1, x2.reshape(mul_ir_in2.dim)) / sqrt(mul_ir_out.ir.dim)
-            elif specialized_code and mul_ir_out.ir.l == 0:
-                out = ins.path_weight * einsum("uvw,ui,vi->w", w, x1, x2) / sqrt(mul_ir_in1.ir.dim)
-            else:
-                out = einsum("uvw,ijk,uvij->wk", w, w3j, xx)
+            out = einsum("uvw,ijk,uvij->wk", w, w3j, xx)
         if ins.connection_mode == "uvu":
             assert mul_ir_in1.mul == mul_ir_out.mul
             if ins.has_weight:
-                if specialized_code and (
-                    mul_ir_in1.ir.l,
-                    mul_ir_in2.ir.l,
-                    mul_ir_out.ir.l,
-                ) == (0, 0, 0):
-                    out = ins.path_weight * einsum(
-                        "uv,u,v->u",
-                        w,
-                        x1.reshape(mul_ir_in1.dim),
-                        x2.reshape(mul_ir_in2.dim),
-                    )
-                elif specialized_code and mul_ir_in1.ir.l == 0:
-                    out = ins.path_weight * einsum("uv,u,vj->uj", w, x1.reshape(mul_ir_in1.dim), x2) / sqrt(mul_ir_out.ir.dim)
-                elif specialized_code and mul_ir_in2.ir.l == 0:
-                    out = ins.path_weight * einsum("uv,ui,v->ui", w, x1, x2.reshape(mul_ir_in2.dim)) / sqrt(mul_ir_out.ir.dim)
-                elif specialized_code and mul_ir_out.ir.l == 0:
-                    out = ins.path_weight * einsum("uv,ui,vi->u", w, x1, x2) / sqrt(mul_ir_in1.ir.dim)
-                else:
-                    out = einsum("uv,ijk,uvij->uk", w, w3j, xx)
+                out = einsum("uv,ijk,uvij->uk", w, w3j, xx)
             else:
                 # not so useful operation because v is summed
                 out = einsum("ijk,uvij->uk", w3j, xx)
         if ins.connection_mode == "uvv":
             assert mul_ir_in2.mul == mul_ir_out.mul
             if ins.has_weight:
-                if specialized_code and (
-                    mul_ir_in1.ir.l,
-                    mul_ir_in2.ir.l,
-                    mul_ir_out.ir.l,
-                ) == (0, 0, 0):
-                    out = ins.path_weight * einsum(
-                        "uv,u,v->v",
-                        w,
-                        x1.reshape(mul_ir_in1.dim),
-                        x2.reshape(mul_ir_in2.dim),
-                    )
-                elif specialized_code and mul_ir_in1.ir.l == 0:
-                    out = ins.path_weight * einsum("uv,u,vj->vj", w, x1.reshape(mul_ir_in1.dim), x2) / sqrt(mul_ir_out.ir.dim)
-                elif specialized_code and mul_ir_in2.ir.l == 0:
-                    out = ins.path_weight * einsum("uv,ui,v->vi", w, x1, x2.reshape(mul_ir_in2.dim)) / sqrt(mul_ir_out.ir.dim)
-                elif specialized_code and mul_ir_out.ir.l == 0:
-                    out = ins.path_weight * einsum("uv,ui,vi->v", w, x1, x2) / sqrt(mul_ir_in1.ir.dim)
-                else:
-                    out = einsum("uv,ijk,uvij->vk", w, w3j, xx)
+                out = einsum("uv,ijk,uvij->vk", w, w3j, xx)
             else:
                 # not so useful operation because u is summed
                 out = einsum("ijk,uvij->vk", w3j, xx)
@@ -601,16 +450,111 @@ def _left_right(
     return IrrepsArray.from_list(self.irreps_out, out, ())
 
 
-@partial(jax.jit, static_argnums=(0,), static_argnames=("optimize_einsums", "custom_einsum_vjp"))
+def _fused_left_right(
+    self: FunctionalTensorProduct,
+    weights_flat: jnp.ndarray,
+    input1: IrrepsArray,
+    input2: IrrepsArray,
+    einsum: Callable,
+) -> IrrepsArray:
+    with jax.ensure_compile_time_eval():
+        num_path = weights_flat.size
+        has_path_with_no_weights = any(not ins.has_weight for ins in self.instructions)
+        i = 0
+
+        if has_path_with_no_weights:
+            num_path += 1
+            i += 1
+
+        big_w3j = jnp.zeros(
+            (
+                num_path,
+                self.irreps_in1.dim,
+                self.irreps_in2.dim,
+                self.irreps_out.dim,
+            )
+        )
+        for ins in self.instructions:
+            mul_ir_in1 = self.irreps_in1[ins.i_in1]
+            mul_ir_in2 = self.irreps_in2[ins.i_in2]
+            mul_ir_out = self.irreps_out[ins.i_out]
+            m1, m2, mo = mul_ir_in1.mul, mul_ir_in2.mul, mul_ir_out.mul
+            d1, d2, do = mul_ir_in1.ir.dim, mul_ir_in2.ir.dim, mul_ir_out.ir.dim
+            s1 = self.irreps_in1[: ins.i_in1].dim
+            s2 = self.irreps_in2[: ins.i_in2].dim
+            so = self.irreps_out[: ins.i_out].dim
+
+            w3j = clebsch_gordan(mul_ir_in1.ir.l, mul_ir_in2.ir.l, mul_ir_out.ir.l)
+
+            def set_w3j(x, i, u, v, w):
+                return x.at[
+                    i,
+                    s1 + u * d1 : s1 + (u + 1) * d1,
+                    s2 + v * d2 : s2 + (v + 1) * d2,
+                    so + w * do : so + (w + 1) * do,
+                ].add(ins.path_weight * w3j)
+
+            if ins.connection_mode == "uvw":
+                assert ins.has_weight
+                for u, v, w in itertools.product(range(m1), range(m2), range(mo)):
+                    big_w3j = set_w3j(big_w3j, i, u, v, w)
+                    i += 1
+            elif ins.connection_mode == "uvu":
+                assert ins.has_weight
+                for u, v in itertools.product(range(m1), range(m2)):
+                    big_w3j = set_w3j(big_w3j, i, u, v, u)
+                    i += 1
+            elif ins.connection_mode == "uvv":
+                assert ins.has_weight
+                for u, v in itertools.product(range(m1), range(m2)):
+                    big_w3j = set_w3j(big_w3j, i, u, v, v)
+                    i += 1
+            elif ins.connection_mode == "uuu":
+                for u in range(m1):
+                    if ins.has_weight:
+                        big_w3j = set_w3j(big_w3j, i, u, u, u)
+                        i += 1
+                    else:
+                        big_w3j = set_w3j(big_w3j, 0, u, u, u)
+            elif ins.connection_mode == "uvuv":
+                for u in range(m1):
+                    for v in range(m2):
+                        if ins.has_weight:
+                            big_w3j = set_w3j(big_w3j, i, u, v, u * m2 + v)
+                            i += 1
+                        else:
+                            big_w3j = set_w3j(big_w3j, 0, u, v, u * m2 + v)
+            else:
+                assert False
+
+        assert i == num_path
+
+    if has_path_with_no_weights and big_w3j.shape[0] == 1:
+        big_w3j = big_w3j.reshape(big_w3j.shape[1:])
+        out = einsum("ijk,i,j->k", big_w3j, input1.array, input2.array)
+    else:
+        if has_path_with_no_weights:
+            weights_flat = jnp.concatenate([jnp.ones((1,)), weights_flat])
+
+        out = einsum(
+            "p,pijk,i,j->k",
+            weights_flat,
+            big_w3j,
+            input1.array,
+            input2.array,
+        )
+    return IrrepsArray(self.irreps_out, out)
+
+
+@partial(jax.jit, static_argnums=(0,), static_argnames=("custom_einsum_jvp"))
 @partial(jax.profiler.annotate_function, name="TensorProduct.right")
 def _right(
     self: FunctionalTensorProduct,
-    weights,
-    input2,
+    weights: List[jnp.ndarray],
+    input2: IrrepsArray,
     *,
-    optimize_einsums=False,
-    custom_einsum_vjp=False,
-):
+    custom_einsum_jvp: bool = False,
+) -> jnp.ndarray:
     # = Short-circut for zero dimensional =
     if self.irreps_in1.dim == 0 or self.irreps_in2.dim == 0 or self.irreps_out.dim == 0:
         return jnp.zeros(
@@ -620,11 +564,7 @@ def _right(
             )
         )
 
-    if custom_einsum_vjp:
-        assert optimize_einsums
-        einsum = opt_einsum
-    else:
-        einsum = partial(jnp.einsum, optimize="optimal" if optimize_einsums else "greedy")
+    einsum = opt_einsum if custom_einsum_jvp else jnp.einsum
 
     weight_index = 0
 
