@@ -1,8 +1,9 @@
 import collections
 import dataclasses
 import itertools
+import warnings
 from functools import partial
-from typing import List
+from typing import List, Tuple, Union
 
 import jax
 import jax.numpy as jnp
@@ -10,6 +11,8 @@ import jax.scipy
 
 from e3nn_jax import axis_angle_to_angles, config, matrix_to_angles, perm, quaternion_to_angles, wigner_D
 
+IntoIrrep = Union[int, "Irrep", "MulIrrep", Tuple[int, int]]
+
 
 @dataclasses.dataclass(init=False, frozen=True)
 class Irrep:
@@ -51,7 +54,7 @@ class Irrep:
     l: int
     p: int
 
-    def __init__(self, l, p=None):
+    def __init__(self, l: IntoIrrep, p=None):
         if p is None:
             if isinstance(l, Irrep):
                 p = l.p
@@ -107,17 +110,17 @@ class Irrep:
         (matrix) Representation of :math:`O(3)`. :math:`D` is the representation of :math:`SO(3)`, see `wigner_D`.
 
         Args:
-            alpha (``jnp.ndarray``): of shape :math:`(...)`
+            alpha (`jax.numpy.ndarray`): of shape :math:`(...)`
                 Rotation :math:`\alpha` around Y axis, applied third.
-            beta (``jnp.ndarray``): of shape :math:`(...)`
+            beta (`jax.numpy.ndarray`): of shape :math:`(...)`
                 Rotation :math:`\beta` around X axis, applied second.
-            gamma (``jnp.ndarray``): of shape :math:`(...)`
+            gamma (`jax.numpy.ndarray`): of shape :math:`(...)`
                 Rotation :math:`\gamma` around Y axis, applied first.
-            k (optional ``jnp.ndarray``): of shape :math:`(...)`
+            k (optional `jax.numpy.ndarray`): of shape :math:`(...)`
                 How many times the parity is applied.
 
         Returns:
-            ``jnp.ndarray``: of shape :math:`(..., 2l+1, 2l+1)`
+            `jax.numpy.ndarray`: of shape :math:`(..., 2l+1, 2l+1)`
 
         See Also:
             o3.wigner_D
@@ -131,11 +134,11 @@ class Irrep:
         r"""Matrix of the representation, see `Irrep.D_from_angles`
 
         Args:
-            q (``jnp.ndarray``): shape :math:`(..., 4)`
-            k (optional ``jnp.ndarray``): shape :math:`(...)`
+            q (`jax.numpy.ndarray`): shape :math:`(..., 4)`
+            k (optional `jax.numpy.ndarray`): shape :math:`(...)`
 
         Returns:
-            ``jnp.ndarray``: shape :math:`(..., 2l+1, 2l+1)`
+            `jax.numpy.ndarray`: shape :math:`(..., 2l+1, 2l+1)`
         """
         return self.D_from_angles(*quaternion_to_angles(q), k)
 
@@ -143,11 +146,11 @@ class Irrep:
         r"""Matrix of the representation
 
         Args:
-            R (``jnp.ndarray``): array of shape :math:`(..., 3, 3)`
-            k (``jnp.ndarray``, optional): array of shape :math:`(...)`
+            R (`jax.numpy.ndarray`): array of shape :math:`(..., 3, 3)`
+            k (`jax.numpy.ndarray`, optional): array of shape :math:`(...)`
 
         Returns:
-            ``jnp.ndarray``: array of shape :math:`(..., 2l+1, 2l+1)`
+            `jax.numpy.ndarray`: array of shape :math:`(..., 2l+1, 2l+1)`
 
         Examples:
             >>> m = Irrep(1, -1).D_from_matrix(-jnp.eye(3))
@@ -244,6 +247,22 @@ class MulIrrep:
 
 jax.tree_util.register_pytree_node(MulIrrep, lambda mulir: ((), mulir), lambda mulir, _: mulir)
 
+IntoIrreps = Union[
+    None,
+    Irrep,
+    MulIrrep,
+    str,
+    "Irreps",
+    List[
+        Union[
+            str,
+            Irrep,
+            MulIrrep,
+            Tuple[int, IntoIrrep],
+        ]
+    ],
+]
+
 
 class Irreps(tuple):
     r"""Direct sum of irreducible representations of :math:`O(3)`
@@ -284,13 +303,17 @@ class Irreps(tuple):
         (, )
     """
 
-    def __new__(cls, irreps=None):
+    def __new__(cls, irreps: IntoIrreps = None):
         if isinstance(irreps, Irreps):
             return super().__new__(cls, irreps)
 
-        out = []
+        out: List[MulIrrep] = []
         if isinstance(irreps, Irrep):
             out.append(MulIrrep(1, Irrep(irreps)))
+        elif irreps is None:
+            pass
+        elif isinstance(irreps, MulIrrep):
+            out.append(irreps)
         elif isinstance(irreps, str):
             try:
                 if irreps.strip() != "":
@@ -307,8 +330,6 @@ class Irreps(tuple):
                         out.append(MulIrrep(mul, ir))
             except Exception:
                 raise ValueError(f'Unable to convert string "{irreps}" into an Irreps')
-        elif irreps is None:
-            pass
         else:
             for mul_ir in irreps:
                 if isinstance(mul_ir, str):
@@ -375,7 +396,7 @@ class Irreps(tuple):
             normalization : {'component', 'norm'}
 
         Returns:
-            ``jnp.ndarray``: array of shape ``size`` where ``-1`` is replaced by ``self.dim``
+            `jax.numpy.ndarray`: array of shape ``size`` where ``-1`` is replaced by ``self.dim``
 
         Examples:
             >>> key = jax.random.PRNGKey(0)
@@ -386,6 +407,7 @@ class Irreps(tuple):
             >>> jnp.max(jnp.abs(jnp.linalg.norm(random_tensor, axis=1) - 1)) < 1e-5
             DeviceArray(True, dtype=bool)
         """
+        warnings.warn("Irreps.randn is deprecated, use e3nn.normal instead", DeprecationWarning)
         di = size.index(-1)
         lsize = size[:di]
         rsize = size[di + 1 :]
@@ -407,7 +429,7 @@ class Irreps(tuple):
         else:
             raise ValueError("Normalization needs to be 'norm' or 'component'")
 
-    def __getitem__(self, i):
+    def __getitem__(self, i) -> Union[MulIrrep, "Irreps"]:
         x = super().__getitem__(i)
         if isinstance(i, slice):
             return Irreps(x)
@@ -589,15 +611,19 @@ class Irreps(tuple):
         r"""Rotate the data by angles according to the irreps
 
         Args:
-            array (``jnp.ndarray``): data compatible with the irreps.
+            array (`jax.numpy.ndarray`): data compatible with the irreps.
             alpha (float): third rotation angle around the second axis (in radians)
             beta (float): second rotation angle around the first axis (in radians)
             gamma (float): first rotation angle around the second axis (in radians)
             k (int): parity operation
 
         Returns:
-            ``jnp.ndarray``
+            `jax.numpy.ndarray`
         """
+        warnings.warn(
+            "Irreps.transform_by_angles is deprecated, use IrrepsArray.transform_by_angles instead", DeprecationWarning
+        )
+
         assert array.shape[-1] == self.dim
         D = self.D_from_angles(alpha, beta, gamma, k)
         return jnp.einsum("ij,...j", D, array)
@@ -607,12 +633,12 @@ class Irreps(tuple):
         r"""Rotate data by a rotation given by a quaternion
 
         Args:
-            array (``jnp.ndarray``): data compatible with the irreps.
-            q (``jnp.ndarray``): quaternion
+            array (`jax.numpy.ndarray`): data compatible with the irreps.
+            q (`jax.numpy.ndarray`): quaternion
             k (int): parity operation
 
         Returns:
-            ``jnp.ndarray``
+            `jax.numpy.ndarray`
 
         Examples:
             >>> irreps = Irreps("0e + 1o + 2e")
@@ -623,6 +649,10 @@ class Irreps(tuple):
             ... ) + 0.0
             DeviceArray([ 1.,  0., -1.,  0.,  0.,  0.,  1.,  0.,  0.], dtype=float32)
         """
+        warnings.warn(
+            "Irreps.transform_by_quaternion is deprecated, use IrrepsArray.transform_by_quaternion instead", DeprecationWarning
+        )
+
         return self.transform_by_angles(array, *quaternion_to_angles(q), k)
 
     @partial(jax.jit, static_argnums=(0,), inline=True)
@@ -630,14 +660,18 @@ class Irreps(tuple):
         r"""Rotate data by an axis and an angle
 
         Args:
-            array (``jnp.ndarray``): data compatible with the irreps.
-            axis (``jnp.ndarray``): axis
+            array (`jax.numpy.ndarray`): data compatible with the irreps.
+            axis (`jax.numpy.ndarray`): axis
             angle (float): angle (in radians)
             k (int): parity operation
 
         Returns:
-            ``jnp.ndarray``
+            `jax.numpy.ndarray`
         """
+        warnings.warn(
+            "Irreps.transform_by_axis_angle is deprecated, use IrrepsArray.transform_by_axis_angle instead", DeprecationWarning
+        )
+
         return self.transform_by_angles(array, *axis_angle_to_angles(axis, angle), k)
 
     @partial(jax.jit, static_argnums=(0,), inline=True)
@@ -645,12 +679,16 @@ class Irreps(tuple):
         r"""Rotate data by a rotation given by a matrix
 
         Args:
-            array (``jnp.ndarray``): data compatible with the irreps.
-            R (``jnp.ndarray``): rotation matrix
+            array (`jax.numpy.ndarray`): data compatible with the irreps.
+            R (`jax.numpy.ndarray`): rotation matrix
 
         Returns:
-            ``jnp.ndarray``
+            `jax.numpy.ndarray`
         """
+        warnings.warn(
+            "Irreps.transform_by_matrix is deprecated, use IrrepsArray.transform_by_matrix instead", DeprecationWarning
+        )
+
         d = jnp.sign(jnp.linalg.det(R))
         R = d[..., None, None] * R
         k = (1 - d) / 2
@@ -667,7 +705,7 @@ class Irreps(tuple):
             k (int): parity operation
 
         Returns:
-            ``jnp.ndarray``: array of shape :math:`(..., \mathrm{dim}, \mathrm{dim})`
+            `jax.numpy.ndarray`: array of shape :math:`(..., \mathrm{dim}, \mathrm{dim})`
         """
         return jax.scipy.linalg.block_diag(*[ir.D_from_angles(alpha, beta, gamma, k) for mul, ir in self for _ in range(mul)])
 
@@ -676,11 +714,11 @@ class Irreps(tuple):
         r"""Matrix of the representation
 
         Args:
-            q (``jnp.ndarray``): array of shape :math:`(..., 4)`
-            k (``jnp.ndarray``, optional): array of shape :math:`(...)`
+            q (`jax.numpy.ndarray`): array of shape :math:`(..., 4)`
+            k (`jax.numpy.ndarray`, optional): array of shape :math:`(...)`
 
         Returns:
-            ``jnp.ndarray``: array of shape :math:`(..., \mathrm{dim}, \mathrm{dim})`
+            `jax.numpy.ndarray`: array of shape :math:`(..., \mathrm{dim}, \mathrm{dim})`
         """
         return self.D_from_angles(*quaternion_to_angles(q), k)
 
@@ -689,10 +727,10 @@ class Irreps(tuple):
         r"""Matrix of the representation
 
         Args:
-            R (``jnp.ndarray``): array of shape :math:`(..., 3, 3)`
+            R (`jax.numpy.ndarray`): array of shape :math:`(..., 3, 3)`
 
         Returns:
-            ``jnp.ndarray``: array of shape :math:`(..., \mathrm{dim}, \mathrm{dim})`
+            `jax.numpy.ndarray`: array of shape :math:`(..., \mathrm{dim}, \mathrm{dim})`
         """
         d = jnp.sign(jnp.linalg.det(R))
         R = d[..., None, None] * R
