@@ -56,13 +56,13 @@ def sh(
 
     Args:
         irreps_out (`Irreps` or int or Sequence[int]): the output irreps
-        input (``jnp.ndarray``): cartesian coordinates
+        input (`jax.numpy.ndarray`): cartesian coordinates
         normalize (bool): if True, the polynomials are restricted to the sphere
         normalization (str): normalization of the constant :math:`\text{cste}`. Default is 'integral'
-        algorithm (Tuple[str]): algorithm to use for the computation. (legendre|recursive, dense|sparse, [custom_vjp])
+        algorithm (Tuple[str]): algorithm to use for the computation. (legendre|recursive, dense|sparse, [custom_jvp])
 
     Returns:
-        ``jnp.ndarray``: polynomials of the spherical harmonics
+        `jax.numpy.ndarray`: polynomials of the spherical harmonics
     """
     input = IrrepsArray("1e", input)
     return spherical_harmonics(irreps_out, input, normalize, normalization, algorithm=algorithm).array
@@ -110,10 +110,10 @@ def spherical_harmonics(
 
     Args:
         irreps_out (`Irreps` or list of int or int): output irreps
-        input (`IrrepsArray` or ``jnp.ndarray``): cartesian coordinates
+        input (`IrrepsArray` or `jax.numpy.ndarray`): cartesian coordinates
         normalize (bool): if True, the polynomials are restricted to the sphere
         normalization (str): normalization of the constant :math:`\text{cste}`. Default is 'integral'
-        algorithm (Tuple[str]): algorithm to use for the computation. (legendre|recursive, dense|sparse, [custom_vjp])
+        algorithm (Tuple[str]): algorithm to use for the computation. (legendre|recursive, dense|sparse, [custom_jvp])
 
     Returns:
         `IrrepsArray`: polynomials of the spherical harmonics
@@ -140,14 +140,15 @@ def spherical_harmonics(
 
     if algorithm is None:
         if config("spherical_harmonics_algorithm") == "automatic":
+            # NOTE the dense algorithm is faster to jit than the sparse one
             if irreps_out.lmax <= 8:
-                algorithm = ("recursive", "sparse", "custom_vjp")
+                algorithm = ("recursive", "dense", "custom_jvp")
             else:
-                algorithm = ("legendre", "sparse", "custom_vjp")
+                algorithm = ("legendre", "dense", "custom_jvp")
         else:
             algorithm = config("spherical_harmonics_algorithm")
 
-    assert all(keyword in ["legendre", "recursive", "dense", "sparse", "custom_vjp"] for keyword in algorithm)
+    assert all(keyword in ["legendre", "recursive", "dense", "sparse", "custom_jvp"] for keyword in algorithm)
 
     if isinstance(input, IrrepsArray):
         [(mul, ir)] = input.irreps
@@ -164,7 +165,7 @@ def spherical_harmonics(
         x = x / jnp.where(r == 0.0, 1.0, r)
 
     sh = _jited_spherical_harmonics(tuple(ir.l for _, ir in irreps_out), x, normalization, algorithm)
-    sh = [jnp.repeat(y[..., None, :], mul, -2) for (mul, ir), y in zip(irreps_out, sh)]
+    sh = [jnp.repeat(y[..., None, :], mul, -2) if mul != 1 else y[..., None, :] for (mul, ir), y in zip(irreps_out, sh)]
     return IrrepsArray.from_list(irreps_out, sh, x.shape[:-1])
 
 
@@ -172,8 +173,8 @@ def spherical_harmonics(
 def _jited_spherical_harmonics(
     ls: Tuple[int, ...], x: jnp.ndarray, normalization: str, algorithm: Tuple[str]
 ) -> List[jnp.ndarray]:
-    if "custom_vjp" in algorithm:
-        return _custom_vjp_spherical_harmonics(ls, x, normalization, algorithm)
+    if "custom_jvp" in algorithm:
+        return _custom_jvp_spherical_harmonics(ls, x, normalization, algorithm)
     else:
         return _spherical_harmonics(ls, x, normalization, algorithm)
 
@@ -191,58 +192,12 @@ def _spherical_harmonics(ls: Tuple[int, ...], x: jnp.ndarray, normalization: str
 
 
 @partial(jax.custom_jvp, nondiff_argnums=(0, 2, 3))
-@partial(jax.custom_vjp, nondiff_argnums=(0, 2, 3))
-def _custom_vjp_spherical_harmonics(
+def _custom_jvp_spherical_harmonics(
     ls: Tuple[int, ...], x: jnp.ndarray, normalization: str, algorithm: Tuple[str]
 ) -> List[jnp.ndarray]:
     return _spherical_harmonics(ls, x, normalization, algorithm)
 
 
-def _fwd(
-    ls: Tuple[int, ...], x: jnp.ndarray, normalization: str, algorithm: Tuple[str]
-) -> Tuple[List[jnp.ndarray], List[jnp.ndarray]]:
-    js = tuple(max(0, l - 1) for l in ls)
-    output = _custom_vjp_spherical_harmonics(ls + js, x, normalization, algorithm)
-
-    return output[: len(ls)], output[len(ls) :]
-
-
-def _bwd(
-    ls: Tuple[int, ...], normalization: str, algorithm: Tuple[str], res: List[jnp.ndarray], grad: List[jnp.ndarray]
-) -> jnp.ndarray:
-    # algo in list with different code per L is faster to execute but very slow to compile!
-    # TODO implement a dense version of this. No list per l. Can use jax.lax.fori_loop
-    # TODO it could be max(ls) dependant. for max(ls) > 10, use a dense algorithm!
-
-    def h(l, r, g):
-        w = clebsch_gordan(l - 1, l, 1)
-        if normalization == "norm":
-            w *= ((2 * l + 1) * l * (2 * l - 1)) ** 0.5
-        else:
-            w *= l**0.5 * (2 * l + 1)
-
-        if "dense" in algorithm:
-            return jnp.einsum("...i,...j,ijk->...k", r, g, w)
-        if "sparse" in algorithm:
-            return jnp.stack(
-                [
-                    sum(
-                        [
-                            w[i, j, k] * r[..., i] * g[..., j]
-                            for i in range(2 * l - 1)
-                            for j in range(2 * l + 1)
-                            if w[i, j, k] != 0
-                        ]
-                    )
-                    for k in range(3)
-                ],
-                axis=-1,
-            )
-        raise ValueError("Unknown algorithm: must be 'dense' or 'sparse'")
-
-    return (sum([h(l, r, g) if l > 0 else jnp.zeros_like(r, shape=r.shape[:-1] + (3,)) for l, r, g in zip(ls, res, grad)]),)
-
-
 def _jvp(
     ls: Tuple[int, ...], normalization: str, algorithm: Tuple[str], primals: Tuple[jnp.ndarray], tangents: Tuple[jnp.ndarray]
 ) -> List[jnp.ndarray]:
@@ -250,7 +205,7 @@ def _jvp(
     (x_dot,) = tangents
 
     js = tuple(max(0, l - 1) for l in ls)
-    output = _custom_vjp_spherical_harmonics(ls + js, x, normalization, algorithm)
+    output = _custom_jvp_spherical_harmonics(ls + js, x, normalization, algorithm)
     out, res = output[: len(ls)], output[len(ls) :]
 
     def h(l: int, r: jnp.ndarray) -> jnp.ndarray:
@@ -277,8 +232,7 @@ def _jvp(
     return out, [h(l, r) if l > 0 else jnp.zeros_like(r) for l, r in zip(ls, res)]
 
 
-_custom_vjp_spherical_harmonics.defvjp(_fwd, _bwd)
-_custom_vjp_spherical_harmonics.defjvp(_jvp)
+_custom_jvp_spherical_harmonics.defjvp(_jvp)
 
 
 def _recursive_spherical_harmonics(
@@ -286,21 +240,26 @@ def _recursive_spherical_harmonics(
 ) -> sympy.Array:
     context.update(dict(jnp=jnp, clebsch_gordan=clebsch_gordan))
 
-    if 0 not in context:
-        if normalization == "integral":
-            context[0] = math.sqrt(1 / (4 * math.pi)) * jnp.ones_like(input[..., :1])
-            context[1] = math.sqrt(3 / (4 * math.pi)) * input
-        elif normalization == "component":
-            context[0] = jnp.ones_like(input[..., :1])
-            context[1] = math.sqrt(3) * input
-        else:
-            context[0] = jnp.ones_like(input[..., :1])
-            context[1] = input
-
     if l == 0:
+        if 0 not in context:
+            if normalization == "integral":
+                context[0] = math.sqrt(1 / (4 * math.pi)) * jnp.ones_like(input[..., :1])
+            elif normalization == "component":
+                context[0] = jnp.ones_like(input[..., :1])
+            else:
+                context[0] = jnp.ones_like(input[..., :1])
+
         return sympy.Array([1])
 
     if l == 1:
+        if 1 not in context:
+            if normalization == "integral":
+                context[1] = math.sqrt(3 / (4 * math.pi)) * input
+            elif normalization == "component":
+                context[1] = math.sqrt(3) * input
+            else:
+                context[1] = input
+
         return sympy.Array([1, 0, 0])
 
     def sh_var(l):
