@@ -1,3 +1,5 @@
+import warnings
+from functools import partial
 from typing import List, Optional
 
 import haiku as hk
@@ -60,15 +62,20 @@ class FullyConnectedTensorProduct(hk.Module):
     def __init__(self, irreps_out, *, irreps_in1=None, irreps_in2=None):
         super().__init__()
 
+        warnings.warn(
+            "e3nn.FullyConnectedTensorProduct is deprecated. Use e3nn.tensor_product followed by e3nn.Linear instead.",
+            DeprecationWarning,
+        )
+
         self.irreps_out = Irreps(irreps_out)
         self.irreps_in1 = Irreps(irreps_in1) if irreps_in1 is not None else None
         self.irreps_in2 = Irreps(irreps_in2) if irreps_in2 is not None else None
 
     def __call__(self, x1: IrrepsArray, x2: IrrepsArray) -> IrrepsArray:
         if self.irreps_in1 is not None:
-            x1 = IrrepsArray.from_any(self.irreps_in1, x1)
+            x1 = x1.convert(self.irreps_in1)
         if self.irreps_in2 is not None:
-            x2 = IrrepsArray.from_any(self.irreps_in2, x2)
+            x2 = x2.convert(self.irreps_in2)
 
         x1 = x1.remove_nones().simplify()
         x2 = x2.remove_nones().simplify()
@@ -90,23 +97,65 @@ class FullyConnectedTensorProduct(hk.Module):
         return output.convert(self.irreps_out)
 
 
-@overload_for_irreps_without_array((0, 1))
 def full_tensor_product(
     input1: IrrepsArray,
     input2: IrrepsArray,
     filter_ir_out: Optional[List[Irrep]] = None,
     irrep_normalization: Optional[str] = None,
 ):
-    r"""Full tensor product of two irreps.
+    warnings.warn("e3nn.full_tensor_product is deprecated. Use e3nn.tensor_product instead.", DeprecationWarning)
+
+    return tensor_product(input1, input2, filter_ir_out=filter_ir_out, irrep_normalization=irrep_normalization)
+
+
+@overload_for_irreps_without_array((0, 1))
+def tensor_product(
+    input1: IrrepsArray,
+    input2: IrrepsArray,
+    *,
+    filter_ir_out: Optional[List[Irrep]] = None,
+    irrep_normalization: Optional[str] = None,
+    custom_einsum_jvp: bool = None,
+    fused: bool = None,
+):
+    """Tensor product of two sets of irreps
 
     Args:
-        input1 (IrrepsArray): First input.
-        input2 (IrrepsArray): Second input.
-        filter_ir_out (Optional[List[Irrep]]): List of irreps to keep in the output.
-        irrep_normalization (Optional[str]): How to normalize the output. See :func:`e3nn_jax.FunctionalTensorProduct`.
+        input1 (IrrepsArray): First input
+        input2 (IrrepsArray): Second input
+        filter_ir_out (list of Irrep, optional): Filter the output irreps. Defaults to None.
+        irrep_normalization (str, optional): Irrep normalization, ``"component"`` or ``"norm"``. Defaults to ``"component"``.
 
     Returns:
-        IrrepsArray: Output with sorted irreps.
+        IrrepsArray: Tensor product of the two inputs.
+            The irreps are sorted (``0e, 0o, 1o, 1e, 2e, 2o, ...``) but not simplified, see example below.
+
+    Examples:
+        >>> jnp.set_printoptions(precision=2, suppress=True)
+        >>> import e3nn_jax as e3nn
+        >>> x = e3nn.IrrepsArray("2x0e + 1o", jnp.arange(5))
+        >>> y = e3nn.IrrepsArray("0o + 2o", jnp.arange(6))
+        >>> e3nn.tensor_product(x, y)
+        2x0o+1x1e+1x1e+1x2e+2x2o+1x3e
+        [  0.     0.     0.     0.     0.    -1.9   16.65  14.83   7.35 -12.57
+           0.    -0.66   4.08   0.     0.     0.     0.     0.     1.     2.
+           3.     4.     5.     9.9   10.97   9.27  -1.97  12.34  15.59  12.73]
+
+        Usage in combination with `Linear`:
+
+        >>> @hk.without_apply_rng
+        ... @hk.transform
+        ... def fully_connected_tensor_product(x, y):
+        ...     return e3nn.Linear("3x1e")(e3nn.tensor_product(x, y))
+        >>> params = fully_connected_tensor_product.init(jax.random.PRNGKey(0), x, y)
+        >>> jax.tree_util.tree_structure(params)
+        PyTreeDef({'linear': {'w[1,0] 2x1e,3x1e': *}})
+        >>> z = fully_connected_tensor_product.apply(params, x, y)
+
+        The irreps can be determined without providing input data:
+
+        >>> e3nn.tensor_product("2x1e + 2e", "2e")
+        1x0e+2x1e+1x1e+2x2e+1x2e+2x3e+1x3e+1x4e
     """
     if filter_ir_out is not None:
         filter_ir_out = [Irrep(ir) for ir in filter_ir_out]
@@ -133,7 +182,7 @@ def full_tensor_product(
         input1.irreps, input2.irreps, irreps_out, instructions, irrep_normalization=irrep_normalization
     )
 
-    return naive_broadcast_decorator(tp.left_right)(input1, input2)
+    return naive_broadcast_decorator(partial(tp.left_right, fused=fused, custom_einsum_jvp=custom_einsum_jvp))(input1, input2)
 
 
 @overload_for_irreps_without_array((0, 1))
@@ -142,12 +191,33 @@ def elementwise_tensor_product(
     input2: IrrepsArray,
     filter_ir_out=None,
     irrep_normalization: str = None,
-    path_normalization: str = None,
 ) -> IrrepsArray:
+    r"""Elementwise tensor product of two `IrrepsArray`
+
+    Args:
+        input1 (IrrepsArray): First input
+        input2 (IrrepsArray): Second input with the same number of irreps as ``input1``,
+            ``input1.irreps.num_irreps == input2.irreps.num_irreps``.
+        filter_ir_out (list of Irrep, optional): Filter the output irreps. Defaults to None.
+        irrep_normalization (str, optional): Irrep normalization, ``"component"`` or ``"norm"``. Defaults to ``"component"``.
+
+    Returns:
+        IrrepsArray: Elementwise tensor product of the two inputs.
+            The irreps are not sorted and not simplified.
+
+    Examples:
+        >>> jnp.set_printoptions(precision=2, suppress=True)
+        >>> import e3nn_jax as e3nn
+        >>> x = e3nn.IrrepsArray("2x0e + 1o", jnp.arange(5))
+        >>> y = e3nn.IrrepsArray("1e + 0o + 0o", jnp.arange(5))
+        >>> e3nn.elementwise_tensor_product(x, y)
+        1x1e+1x0o+1x1e [ 0.  0.  0.  3.  8. 12. 16.]
+    """
     if filter_ir_out is not None:
         filter_ir_out = [Irrep(ir) for ir in filter_ir_out]
 
-    assert input1.irreps.num_irreps == input2.irreps.num_irreps
+    if input1.irreps.num_irreps != input2.irreps.num_irreps:
+        raise ValueError(f"Number of irreps must be the same, got {input1.irreps.num_irreps} and {input2.irreps.num_irreps}")
 
     irreps_in1 = list(input1.irreps)
     irreps_in2 = list(input2.irreps)
@@ -188,7 +258,6 @@ def elementwise_tensor_product(
         irreps_out,
         instructions,
         irrep_normalization=irrep_normalization,
-        path_normalization=path_normalization,
     )
 
     return naive_broadcast_decorator(tp.left_right)(input1, input2)
@@ -257,7 +326,7 @@ class TensorSquare(hk.Module):
 
     def __call__(self, input: IrrepsArray) -> IrrepsArray:
         if self.irreps_in is not None:
-            input = IrrepsArray.from_any(self.irreps_in, input)
+            input = input.convert(self.irreps_in)
 
         input = input.remove_nones().simplify()
 
